#!/usr/bin/env python
# pylint:disable=R,C

"""Integration tests using YORM as a persistence model."""

import os

import yorm


class Config:

    def __init__(self, key, name=None, root=None):
        self.key = key
        self.name = name or ""
        self.root = root or ""


@yorm.attr(key=yorm.converters.String)
@yorm.attr(name=yorm.converters.String)
@yorm.sync("{self.root}/{self.key}/config.yml")
class ConfigModel:

    def __init__(self, key, root):
        self.key = key
        self.root = root
        print(self.key)

    @staticmethod
    def pm_to_dm(model):
        config = Config(model.key)
        config.name = model.name
        config.root = model.root
        return config


class ConfigStore:

    def __init__(self, root):
        self.root = root
        self.path = self.root + "/{}/config.yml"
        self.attrs = dict(key=yorm.converters.String,
                          name=yorm.converters.String)

    def read(self, key):
        instance = Config(key)
        path = self.path.format(key)
        attrs = self.attrs
        try:
            yorm.sync(instance, path, attrs, existing=True, auto=False)
        except yorm.exceptions.FileMissingError:
            return None
        else:
            yorm.update_object(instance)
            return instance


class TestPersistanceMapping:

    root = os.path.join(os.path.dirname(__file__), 'files')

    def test_load_pm(self):
        model = ConfigModel('my_key', self.root)

        print(model.__dict__)
        assert model.key == "my_key"
        assert model.root == self.root
        assert model.name == "my_name"

    def test_create_dm_from_pm(self):
        model = ConfigModel('my_key', self.root)
        config = ConfigModel.pm_to_dm(model)

        print(config.__dict__)
        assert config.key == "my_key"
        assert config.root == self.root
        assert config.name == "my_name"


class TestStore:

    def test_read_missing(self, tmpdir):
        store = ConfigStore(str(tmpdir))
        assert None is store.read('unknown')



"""Custom exceptions."""

from abc import ABCMeta

import yaml


class Error(Exception, metaclass=ABCMeta):

    """Base class for all YORM exceptions."""


class FileMissingError(Error, FileNotFoundError):

    """A file was expected to exist."""


class FileAlreadyExistsError(Error, FileExistsError):

    """A file was not expected to exist."""


class FileDeletedError(Error, FileNotFoundError):

    """Text could not be read from a deleted file."""


class ContentError(Error, yaml.error.YAMLError, ValueError):

    """Text could not be parsed as valid YAML."""


class ConversionError(Error, ValueError):

    """Value could not be converted to the specified type."""


class MappingError(Error):

    """The API was called incorrectly."""



# pylint:disable=R,C

"""Integration tests for file IO."""

import logging

import pytest

from yorm.common import write_text

from . import refresh_file_modification_times
from .samples import *  # pylint: disable=W0401,W0614


class TestCreate:

    """Integration tests for creating mapped classes."""

    def test_fetch_from_existing(self, tmpdir):
        """Verify attributes are updated from an existing file."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample2 = SampleStandardDecorated('sample')
        assert sample2.yorm_mapper.path == sample.yorm_mapper.path

        refresh_file_modification_times()

        logging.info("changing values in object 1...")
        sample.array = [0, 1, 2]
        sample.string = "Hello, world!"
        sample.number_int = 42
        sample.number_real = 4.2
        sample.true = True
        sample.false = False

        logging.info("reading changed values in object 2...")
        assert [0, 1, 2] == sample2.array
        assert "Hello, world!" == sample2.string
        assert 42 == sample2.number_int
        assert 4.2 == sample2.number_real
        assert True is sample2.true
        assert False is sample2.false


class TestDelete:

    """Integration tests for deleting files."""

    def test_read(self, tmpdir):
        """Verify a deleted file cannot be read from."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample.yorm_mapper.delete()

        with pytest.raises(FileNotFoundError):
            print(sample.string)

        with pytest.raises(FileNotFoundError):
            sample.string = "def456"

    def test_write(self, tmpdir):
        """Verify a deleted file cannot be written to."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample.yorm_mapper.delete()

        with pytest.raises(FileNotFoundError):
            sample.string = "def456"

    def test_multiple(self, tmpdir):
        """Verify a deleted file can be deleted again."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample.yorm_mapper.delete()
        sample.yorm_mapper.delete()


class TestUpdate:

    """Integration tests for updating files/object."""

    def test_automatic_store_after_first_modification(self, tmpdir):
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        assert "number_int: 0\n" in sample.yorm_mapper.text

        sample.number_int = 42
        assert "number_int: 42\n" in sample.yorm_mapper.text

        sample.yorm_mapper.text = "number_int: true\n"
        assert 1 is sample.number_int
        assert "number_int: 1\n" in sample.yorm_mapper.text

    def test_automatic_store_after_first_modification_on_list(self, tmpdir):
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        assert "array: []\n" in sample.yorm_mapper.text

        sample.array.append(42)
        assert "array:\n- 42\n" in sample.yorm_mapper.text

        sample.yorm_mapper.text = "array: [true]\n"
        try:
            iter(sample)
        except AttributeError:
            pass
        assert "array:\n- 1\n" in sample.yorm_mapper.text



"""Functions and decorators."""

import uuid

from . import common, exceptions
from .bases import Mappable
from .mapper import get_mapper, set_mapper

log = common.logger(__name__)

UUID = 'UUID'


def sync(*args, **kwargs):
    """Convenience function to forward calls based on arguments.

    This function will call either:

    * `sync_object` - when given an unmapped object
    * `sync_instances` - when used as the class decorator

    Consult the signature of each call for more information.

    """
    if 'path_format' in kwargs or args and isinstance(args[0], str):
        return sync_instances(*args, **kwargs)
    else:
        return sync_object(*args, **kwargs)


def sync_object(instance, path, attrs=None, existing=None, auto=True):
    """Enable YAML mapping on an object.

    :param instance: object to patch with YAML mapping behavior
    :param path: file path for dump/load
    :param attrs: dictionary of attribute names mapped to converter classes
    :param existing: indicate if file is expected to exist or not
    :param auto: automatically store attributes to file

    """
    log.info("mapping %r to %s...", instance, path)
    _check_base(instance, mappable=False)

    attrs = attrs or common.attrs[instance.__class__]

    class Mapped(Mappable, instance.__class__):

        """Original class with `Mappable` as the base."""

    instance.__class__ = Mapped

    mapper = set_mapper(instance, path, attrs, auto=auto)
    _check_existance(mapper, existing)

    if mapper.auto:
        if not mapper.exists:
            mapper.create()
            mapper.store()
        mapper.fetch()

    log.info("mapped %r to '%s'", instance, path)
    return instance


def sync_instances(path_format, format_spec=None, attrs=None, **kwargs):
    """Class decorator to enable YAML mapping after instantiation.

    :param path_format: formatting string to create file paths for dump/load
    :param format_spec: dictionary to use for string formatting
    :param attrs: dictionary of attribute names mapped to converter classes
    :param existing: indicate if file is expected to exist or not
    :param auto: automatically store attribute to file

    """
    format_spec = format_spec or {}
    attrs = attrs or {}

    def decorator(cls):
        """Class decorator to map instances to files.."""

        old_init = cls.__init__

        def new_init(self, *_args, **_kwargs):
            """Modified class __init__ that maps the resulting instance."""
            old_init(self, *_args, **_kwargs)

            log.info("mapping instance of %r to '%s'...", cls, path_format)

            format_values = {}
            for key, value in format_spec.items():
                format_values[key] = getattr(self, value)
            if '{' + UUID + '}' in path_format:
                format_values[UUID] = uuid.uuid4().hex
            format_values['self'] = self

            path = path_format.format(**format_values)
            attrs.update(common.attrs[self.__class__])
            attrs.update(common.attrs[cls])

            sync_object(self, path, attrs, **kwargs)

        new_init.__doc__ = old_init.__doc__
        cls.__init__ = new_init

        return cls

    return decorator


def attr(**kwargs):
    """Class decorator to map attributes to converters.

    :param kwargs: keyword arguments mapping attribute name to converter class

    """
    def decorator(cls):
        """Class decorator."""
        for name, converter in kwargs.items():
            common.attrs[cls][name] = converter

        return cls

    return decorator


def update(instance, fetch=True, force=True, store=True):
    """Synchronize changes between a mapped object and its file.

    :param instance: object with patched YAML mapping behavior
    :param fetch: update the object with changes from its file
    :param force: even if the file appears unchanged
    :param store: update the file with changes from the object

    """
    _check_base(instance, mappable=True)

    if fetch:
        update_object(instance, force=False)
    if store:
        update_file(instance)
    if fetch:
        update_object(instance, force=force)


def update_object(instance, existing=True, force=True):
    """Synchronize changes into a mapped object from its file.

    :param instance: object with patched YAML mapping behavior
    :param existing: indicate if file is expected to exist or not
    :param force: update the object even if the file appears unchanged

    """
    log.info("manually updating %r from file...", instance)
    _check_base(instance, mappable=True)

    mapper = get_mapper(instance)
    _check_existance(mapper, existing)

    if mapper.modified or force:
        mapper.fetch()


def update_file(instance, existing=None, force=True):
    """Synchronize changes into a mapped object's file.

    :param instance: object with patched YAML mapping behavior
    :param existing: indicate if file is expected to exist or not
    :param force: update the file even if automatic sync is off

    """
    log.info("manually saving %r to file...", instance)
    _check_base(instance, mappable=True)

    mapper = get_mapper(instance)
    _check_existance(mapper, existing)

    if mapper.auto or force:
        if not mapper.exists:
            mapper.create()
        mapper.store()


def _check_base(obj, mappable=True):
    """Confirm an object's base class is `Mappable` as required."""
    if mappable and not isinstance(obj, Mappable):
        raise exceptions.MappingError("{} is not mapped".format(repr(obj)))
    if not mappable and isinstance(obj, Mappable):
        raise exceptions.MappingError("{} is already mapped".format(repr(obj)))


def _check_existance(mapper, existing=None):
    """Confirm the expected state of the file.

    :param existing: indicate if file is expected to exist or not

    """
    if existing is True:
        if not mapper.exists:
            raise exceptions.FileMissingError
    elif existing is False:
        if mapper.exists:
            raise exceptions.FileAlreadyExistsError
